"""Class for solar calculations."""
from __future__ import division

try:
    range = xrange
except NameError:
    pass

import math


class SolarCalcs(object):
    """Solar calculations.

    Args:
        UCM: Urban Canopy - Building Energy Model object.
        BEM: Building Energy Model object.
        simTime: Simulation time object
        RSM: Rural Site & Vertical Diffusion Model Object.
        forc: Forcing object.
        parameter: Param Object.
        rural: Rural road Element object.

    Properties:
        * # Properties added from solarcalcs
        * dir -- direct sunlight (perpendicular to the sun's ray)
        * dif -- diffuse sunlight
        * tanzen -- tangente of solar zenithal angle
        * critOrient -- critical canyon angle for which solar radiation reaches the road
        * horSol -- direct horizontal radiation
        * Kw_term -- fractional terms for wall
        * Kr_term -- fractional terms for road
        * mr -- fractional terms for wall
        * mw -- fractional terms for road
        * ut -- elapsed hours on current day
        * ad -- fractional year in radians
        * eqtime -- time variable
        * decsol -- solar declination angle
        * zenith -- Angle between normal to earth's surface and sun position
    """

    def __init__(self, UCM, BEM, simTime, RSM, forc, parameter, rural):
        self.UCM = UCM
        self.BEM = BEM
        self.simTime = simTime
        self.RSM = RSM
        self.forc = forc
        self.parameter = parameter
        self.rural = rural

    def solarcalcs(self):
        """ Solar Calculation.

        Mutates RSM, BEM, and UCM objects based on following parameters:

        * UCM -- Urban Canopy - Building Energy Model object
        * BEM -- Building Energy Model object
        * simTime -- Simulation time object
        * RSM -- Rural Site & Vertical Diffusion Model Object
        * forc -- Forcing object
        * parameter -- Geo Param Object
        * rural -- Rural road Element object
        """

        # Direct sunlight (perpendicular to the sun's ray)
        self.dir = self.forc.dir
        self.dif = self.forc.dif  # Diffuse sunlight

        if (self.dir + self.dif) > 0.:

            # calculate zenith tangent, and critOrient solar angles
            self.solarangles()

            # Direct horizontal radiation
            self.horSol = max(math.cos(self.zenith) * self.dir, 0.0)
            # Fractional terms for wall & road
            self.Kw_term = (
                min(abs(1. / self.UCM.canAspect*(0.5 - self.critOrient / math.pi) + 1 /
                        math.pi * self.tanzen * (1 - math.cos(self.critOrient))), 1.))
            self.Kr_term = (
                min(abs(2. * self.critOrient / math.pi -
                        (2 / math.pi * self.UCM.canAspect * self.tanzen) *
                        (1 - math.cos(self.critOrient))),
                    1 - 2 * self.UCM.canAspect * self.Kw_term))

            # Direct and diffuse solar radiation
            # Assume trees are shorter than buildings
            self.bldSol = self.horSol * self.Kw_term + self.UCM.wallConf * self.dif
            self.roadSol = self.horSol * self.Kr_term + self.UCM.roadConf * self.dif

            # Solar reflections. Add diffuse radiation from vegetation to alb_road if
            # in season
            if self.simTime.month < self.parameter.vegStart or \
               self.simTime.month > self.parameter.vegEnd:
                alb_road = self.UCM.road.albedo
            else:
                alb_road = self.UCM.road.albedo * (1. - self.UCM.road.vegcoverage) + \
                    self.parameter.vegAlbedo * self.UCM.road.vegcoverage

            # First set of reflections
            rr = alb_road * self.roadSol
            rw = self.UCM.alb_wall * self.bldSol

            # bounces
            fr = (
                1. - (1. - 2. * self.UCM.wallConf) * self.UCM.alb_wall +
                (1. - self.UCM.roadConf) * self.UCM.wallConf * alb_road *
                self.UCM.alb_wall)

            # (1.0-self.UCM.roadConf) road to wall view
            self.mr = (rr + (1.0 - self.UCM.roadConf) * alb_road *
                       (rw + self.UCM.wallConf * self.UCM.alb_wall * rr)) / fr
            self.mw = (rw + self.UCM.wallConf * self.UCM.alb_wall * rr) / fr

            # Receiving solar, including bounces (W m-2)
            # Includes road covered by vegetation.
            self.UCM.road.solRec = self.roadSol + \
                (1 - self.UCM.roadConf) * self.mw

            for j in range(len(self.BEM)):
                self.BEM[j].roof.solRec = self.horSol + self.dif
                self.BEM[j].wall.solRec = (
                    self.bldSol + (1 - 2*self.UCM.wallConf)*self.mw +
                    self.UCM.wallConf * self.mr)

            self.rural.solRec = self.horSol + self.dif  # Solar received by rural
            self.UCM.SolRecRoof = self.horSol + self.dif  # Solar received by roof
            self.UCM.SolRecRoad = self.UCM.road.solRec  # Solar received by road
            # Solar received by wall
            self.UCM.SolRecWall = (self.bldSol + (1 - 2 * self.UCM.wallConf) *
                                   self.UCM.road.albedo * self.roadSol)

            # Modification from UWG_Matlab on 09/20:
            # Consolidate and incorporate the treeCoverage fraction into the tree
            # sensible and latent heat (W-m2) calculation. Previously this was
            # factored (inconsistently) at the urbflux and UCMDef.UCModel functions.

            # Vegetation heat (tree) (per m^2 of veg)
            self.UCM.treeSensHeat = (
                (1 - self.parameter.vegAlbedo) * (1 - self.parameter.treeFLat) *
                self.UCM.SolRecRoad * self.UCM.treeCoverage)
            self.UCM.treeLatHeat = (
                (1 - self.parameter.vegAlbedo) * self.parameter.treeFLat *
                self.UCM.SolRecRoad * self.UCM.treeCoverage)

            # Modification from UWG_Matlab on 09/20:
            # Add the sensible and latent heat fraction of grass (vegetation not
            # accounted for by tree fraction). Note that the self.UCM.road.sens property
            # also contains a calculation of sensible heat from vegetation and surface
            # convection, but was not used here because the UCModel function in the
            # UCMDef module already accounts for the entire road convective heat transfer
            # so using the UCM.road.sens property would double-count convection. This
            # sensible heat therefore just accounts for the absorbed solar radiation
            # split into it's sensible heat fraction.

            # Vegetation heat (grass) (per m^2 of veg)
            grasscover = self.UCM.vegcover - self.UCM.treeCoverage
            self.UCM.treeSensHeat += (
                (1 - self.parameter.vegAlbedo) * (1 - self.parameter.grassFLat) *
                self.UCM.SolRecRoad * grasscover)
            self.UCM.treeLatHeat += (
                (1 - self.parameter.vegAlbedo) * self.parameter.grassFLat *
                self.UCM.SolRecRoad * grasscover)

        else:  # No Sun

            self.UCM.road.solRec = 0.
            self.rural.solRec = 0.

            for j in range(len(self.BEM)):
                self.BEM[j].roof.solRec = 0.
                self.BEM[j].wall.solRec = 0.

            self.UCM.SolRecRoad = 0.  # Solar received by road
            self.UCM.SolRecRoof = 0.  # Solar received by roof
            self.UCM.SolRecWall = 0.  # Solar received by wall
            self.UCM.treeSensHeat = 0.
            self.UCM.treeLatHeat = 0.

        return self.rural, self.UCM, self.BEM

    def solarangles(self):
        """Solves for zenith angle, tangent of zenithal angle, and critical canyon angle.

        Calculation based on NOAA from:

        * canAspect -- aspect Ratio of canyon
        * simTime -- simulation time parameters
        * RSM.lon -- longitude (deg)
        * RSM.lat -- latitude (deg)
        * RSM.GMT -- GMT hour correction
        """

        month = self.simTime.month
        day = self.simTime.day
        secDay = self.simTime.secDay  # Total elapsed seconds in simulation
        # inobis: total days for first of month
        # i.e [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
        inobis = self.simTime.inobis

        canAspect = self.UCM.canAspect
        lon = self.RSM.lon
        lat = self.RSM.lat
        GMT = self.RSM.gmt

        # Get elapsed hours on current day
        self.ut = (24. + (int(secDay) / 3600. % 24.)) % 24.
        ibis = list(range(len(inobis)))

        for JI in range(1, 12):
            ibis[JI] = inobis[JI] + 1

        date = day + inobis[month-1] - 1  # Julian day of the year
        # divide circle by 365 days, multiply by elapsed days + hours
        # ad: fractional year (radians)
        self.ad = 2.0 * math.pi / 365. * (date - 1 + (self.ut - (12 / 24.)))

        self.eqtime = (
            229.18 * (0.000075 + 0.001868 * math.cos(self.ad) - 0.032077 *
                      math.sin(self.ad) - 0.01461 * math.cos(2 * self.ad) - 0.040849 *
                      math.sin(2 * self.ad)))

        # Declination angle (angle of sun with equatorial plane)
        self.decsol = (
            0.006918 - 0.399912*math.cos(self.ad) + 0.070257*math.sin(self.ad) -
            0.006758 * math.cos(2. * self.ad) + 0.000907 * math.sin(2. * self.ad) -
            0.002697 * math.cos(3. * self.ad) + 0.00148 * math.sin(3. * self.ad))

        time_offset = self.eqtime - 4. * lon + 60 * GMT
        tst = secDay + time_offset * 60

        ha = (tst / 4. / 60. - 180.) * math.pi / 180.
        zlat = lat * (math.pi / 180.)  # change angle units to radians

        # Calculate zenith solar angle
        self.zenith = (
            math.acos(math.sin(zlat) * math.sin(self.decsol) + math.cos(zlat) *
                      math.cos(self.decsol) * math.cos(ha)))

        # tangente of solar zenithal angle
        if abs(0.5 * math.pi - self.zenith) < 1e-6:
            if 0.5 * math.pi - self.zenith > 0.:
                self.tanzen = math.tan(0.5 * math.pi - 1e-6)

            elif 0.5 * math.pi - self.zenith <= 0.:
                self.tanzen = math.tan(0.5 * math.pi + 1e-6)

        elif abs(self.zenith) < 1e-6:
            # lim x->0 tan(x) -> 0 which results in division by zero error
            # when calculating the critical canyon angle
            # so set tanzen to 1e-6 which will result in critical canyon angle = 90
            self.tanzen = 1e-6

        else:
            self.tanzen = math.tan(self.zenith)

        # critical canyon angle for which solar radiation reaches the road
        # has to do with street canyon orientation for given solar angle
        self.critOrient = math.asin(min(abs(1. / self.tanzen) / canAspect, 1.))



"""Tests for SolarCalcs object."""

import pytest
from .test_base import auto_setup_uwg, setup_open_matlab_ref
from uwg import SolarCalcs


def test_solarangles():
    """Test solar angles."""

    model = auto_setup_uwg()
    model.generate()

    solar = SolarCalcs(model.UCM, model.BEM, model.simTime, model.RSM,
                       model.forc, model.geoParam, model.rural)

    # Timestep every 5 minutes (300s)
    for i in range(int(12 * 24 * 1.5)):  # 1.5 days, Jan 2nd, 12 noon
        solar.simTime.update_date()
    solar.solarangles()

    # test simtime for solar after 1.5 days
    assert solar.ut == pytest.approx(12.0, abs=1e-15)
    assert solar.simTime.day == pytest.approx(2.0, abs=1e-15)
    assert solar.ad == pytest.approx(0.197963373, abs=1e-8)

    # Recalculate solar angles with new time simulation, add to previous time increment
    # Increment time to 22 days, 13hrs, 30min = Jan 22 at 1330
    for i in range(12 * 24 * 20 + 12 * 1 + 6):
        solar.simTime.update_date()

    # Run simulation
    solar.solarangles()

    uwg_python_val = [
        solar.simTime.month,
        solar.simTime.secDay,
        solar.simTime.day,
        solar.simTime.hourDay,
        solar.ut,
        solar.ad,
        solar.eqtime,
        solar.decsol,
        solar.zenith,
        solar.tanzen,
        solar.critOrient]

    # open matlab ref file
    uwg_matlab_val = \
        setup_open_matlab_ref('matlab_solarcalcs', 'matlab_ref_solarangles.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)
    for i in range(len(uwg_matlab_val)):
        assert uwg_python_val[i] == \
            pytest.approx(uwg_matlab_val[i], abs=1e-15), 'error at index={}'.format(i)


def test_solarcalcs():
    """Test solar calculation."""

    model = auto_setup_uwg()
    model.generate()

    # We subtract 11 hours from total timestep so
    # we can stop simulation while we still have sun!
    # New time: Jan 31, 1300
    model.simTime.nt -= 12 * 11

    model.simulate()

    # check date
    assert model.simTime.month == 1
    assert model.simTime.day == 31
    assert model.simTime.secDay / 3600. == pytest.approx(13.0, abs=1e-15)

    # Manual checking

    # Total Radiation (W m-2) at 13:00 on Jan 31 w/o reflected sunlight
    assert model.solar.bldSol == pytest.approx(170.4189961148073, abs=1e-12)
    assert model.solar.roadSol == pytest.approx(307.4032644249752, abs=1e-12)

    # Matlab Checking
    uwg_matlab_val = setup_open_matlab_ref(
        'matlab_solarcalcs', 'matlab_ref_solarcalcs.txt')

    # Modify treeSensHeat based on changes to veg/tree heat balance 10/20
    # grassSensHeat = (
    #     (1 - model.geoParam.vegAlbedo) * (1 - model.geoParam.grassFLat) *
    #     model.UCM.SolRecRoad * (model.grasscover / (1 - model.blddensity)))
    # vegSensHeat = \
    #     (model.solar.UCM.treeSensHeat - grassSensHeat) / model.UCM.treeCoverage
    treeSensHeat = (
        (1 - model.solar.parameter.vegAlbedo) * (1 - model.solar.parameter.treeFLat) *
        model.solar.UCM.SolRecRoad)
    treeLatHeat = (
        (1 - model.solar.parameter.vegAlbedo) * model.solar.parameter.treeFLat *
        model.solar.UCM.SolRecRoad)

    uwg_python_val = [
        model.solar.horSol,  # 0
        model.solar.Kw_term,  # 1
        model.solar.Kr_term,  # 2
        model.solar.mr,  # 3
        model.solar.mw,  # 4
        model.solar.BEM[0].roof.solRec,  # 5
        model.solar.BEM[0].wall.solRec,  # 6
        model.solar.BEM[1].roof.solRec,  # 7
        model.solar.BEM[1].wall.solRec,  # 8
        model.solar.rural.solRec,  # 9
        model.solar.UCM.SolRecRoof,  # 10
        model.solar.UCM.SolRecRoad,  # 11
        model.solar.UCM.SolRecWall,  # 12
        treeSensHeat,  # 13
        treeLatHeat]  # 14

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)
    for i in range(len(uwg_matlab_val)):
        assert uwg_python_val[i] == \
            pytest.approx(uwg_matlab_val[i], abs=1e-15), 'error at index={}'.format(i)



"""Material class"""

from .utilities import float_in_range_excl_incl

try:
    str = basestring
except NameError:
    pass


class Material(object):
    """Material class.

    Args:
        thermalcond: Number for thermal conductivity [W m-1 K-1].
        volheat: Number for volumetric heat capacity [J m-3 K-1].
        name: Text string for name of the Material.

    Properties:
        * thermalcond
        * volheat
    """

    def __init__(self, thermalcond, volheat, name):
        self._name = name
        self.thermalcond = thermalcond
        self.volheat = volheat

    @property
    def name(self):
        """Get or set text string for name of Material."""
        return self._name

    @property
    def thermalcond(self):
        """Get or set number for thermal conductivity [W/(m-K)]."""
        return self._thermalcond

    @thermalcond.setter
    def thermalcond(self, value):
        self._thermalcond = \
            float_in_range_excl_incl(value, mi=0, input_name='thermalcond')

    @property
    def volheat(self):
        """Get or set number for volumetric capacity [J/(m3-K)]."""
        return self._volheat

    @volheat.setter
    def volheat(self, value):
        self._volheat = float_in_range_excl_incl(
            value, mi=0, input_name='volheat')

    @classmethod
    def from_dict(cls, data):
        """Create a Material object from a dictionary.

        Args:
            data: A Material dictionary following the format below.

        .. code-block:: python

            {
            "type": "Material",
            "name": "Concrete",
            "thermalcond": 1.311,  # thermal conductivity [W m-1 K-1]
            "volheat": 1874432.0  # volumetric heat capacity [J m-3 K-1]
            }
        """
        assert data['type'] == 'Material', 'Expected ' \
            'Material dictionary. Got {}.'.format(data['type'])

        return cls(data['thermalcond'], data['volheat'], data['name'])

    def to_dict(self):
        """Material dictionary representation."""
        base = {'type': 'Material'}
        base['name'] = self.name
        base['thermalcond'] = self.thermalcond
        base['volheat'] = self.volheat
        return base

    def __repr__(self):
        return "Material, name: {}\n thermalcond: {}\n volheat: {}".format(
            self.name, self.thermalcond, self.volheat)



"""Function for surface heat flux calculations."""
from __future__ import division

try:
    range = xrange
except NameError:
    pass

from .infracalcs import infracalcs
from math import log


def urbflux(UCM, UBL, BEM, forc, parameter, simTime, RSM):
    """Calculate the surface heat fluxes.

    Args:
        UCM: UCMDef object.
        UBL: UBLDef object.
        BEM: List of BEMDef objects.
        forc: Forcing object.
        parameter: Param object.
        simTime: SimParam object.
        RSM: RSMDef object.

    Returns:
        Tuple with updated UCMDef, UBLDef, and BEM objects.
    """

    T_can = UCM.canTemp
    Cp = parameter.cp
    UCM.Q_roof = 0.
    sigma = 5.67e-8  # Stephan-Boltzman constant
    UCM.roofTemp = 0.  # Average urban roof temperature
    UCM.wallTemp = 0.  # Average urban wall temperature

    for j in range(len(BEM)):
        # Building energy model
        BEM[j].building.BEMCalc(UCM, BEM[j], forc, parameter, simTime)
        BEM[j].ElecTotal = BEM[j].building.ElecTotal * BEM[j].fl_area  # W m-2

        # Update roof infra calc
        e_roof = BEM[j].roof.emissivity
        T_roof = BEM[j].roof.layerTemp[0]
        BEM[j].roof.infra = e_roof * (forc.infra - sigma * T_roof ** 4.)

        # update wall infra calc (road done later)
        e_wall = BEM[j].wall.emissivity
        T_wall = BEM[j].wall.layerTemp[0]
        # calculates the infrared radiation for wall, taking into account radiation
        # exchange from road
        _infra_road_, BEM[j].wall.infra = \
            infracalcs(UCM, forc, UCM.road.emissivity, e_wall, UCM.roadTemp, T_wall)

        # Update element temperatures
        BEM[j].mass.layerTemp = \
            BEM[j].mass.Conduction(
                simTime.dt, BEM[j].building.fluxMass, 1., 0., BEM[j].building.fluxMass)
        BEM[j].roof.SurfFlux(
            forc, parameter, simTime, UCM.canHum, T_can, max(forc.wind, UCM.canWind), 1.,
            BEM[j].building.fluxRoof)
        BEM[j].wall.SurfFlux(
            forc, parameter, simTime, UCM.canHum, T_can, UCM.canWind, 1.,
            BEM[j].building.fluxWall)

        # Note the average wall & roof temperature
        UCM.wallTemp = UCM.wallTemp + BEM[j].frac * BEM[j].wall.layerTemp[0]
        UCM.roofTemp = UCM.roofTemp + BEM[j].frac * BEM[j].roof.layerTemp[0]

    # Update road infra calc (assume walls have similar emissivity, so use the last one)
    UCM.road.infra, _wall_infra = \
        infracalcs(UCM, forc, UCM.road.emissivity, e_wall, UCM.roadTemp, UCM.wallTemp)
    UCM.road.SurfFlux(forc, parameter, simTime, UCM.canHum, T_can, UCM.canWind, 2., 0.)
    UCM.roadTemp = UCM.road.layerTemp[0]

    # Sensible & latent heat flux (total)
    try:
        UCM.latHeat += \
            UCM.latAnthrop + UCM.treeLatHeat + UCM.road.lat * (1. - UCM.bldDensity)
    except TypeError:
        pass  # latheat is None

    # ---------------------------------------------------------------------
    # Advective heat flux to UBL from VDM
    #
    # Note: UWG_Matlab code here is modified to compensate for rounding errors
    # that occur when recursively adding forDens, intAdv1, and intAdv2.
    # This causes issues in the UBL.advHeat calculation when large (1e5)
    # numbers are subtracted to produce small numbers (1e-10) that can
    # differ from equivalent matlab calculations by a factor of 2.
    # Values this small are ~ 0, but for consistency's sake Kahan Summation
    # algorithm is applied to keep margin of difference from UWG_Matlab low.
    # ---------------------------------------------------------------------

    forDens = 0.0
    intAdv1 = 0.0
    intAdv2 = 0.0

    # c1 & c2 stores values truncated by floating point rounding for values < 10^-16
    c1 = 0.0
    c2 = 0.0
    c3 = 0.0

    for iz in range(RSM.nzfor):
        # At c loss of precision at at low order of magnitude, that we need in
        # UBL.advHeat calc. Algebraically t is 0, but with floating pt numbers
        # c will accumulate truncated values
        y = (RSM.densityProfC[iz] * RSM.dz[iz] / (RSM.z[RSM.nzfor - 1] +
             RSM.dz[RSM.nzfor-1] / 2.))
        t = forDens + y
        c1 += (t - forDens) - y
        forDens = t

        y = RSM.windProf[iz] * RSM.tempProf[iz] * RSM.dz[iz]
        t = intAdv1 + y
        c2 += (t - intAdv1) - y
        intAdv1 = t

        y = RSM.windProf[iz] * RSM.dz[iz]
        t = intAdv2 + y
        c3 += (t - intAdv2) - y
        intAdv2 = t

    # Add the truncated values back
    forDens -= c1
    intAdv1 -= c2
    intAdv2 -= c3
    UBL.advHeat = (
        UBL.paralLength * Cp * forDens * (intAdv1 - (UBL.ublTemp * intAdv2)) /
        UBL.urbArea)

    # ---------------------------------------------------------------------
    # Convective heat flux to UBL from UCM (see Appendix - Bueno (2014))
    # ---------------------------------------------------------------------
    zrUrb = 2 * UCM.bldHeight
    zref = RSM.z[RSM.nzref-1]  # Reference height

    # Reference wind speed & canyon air density
    windUrb = (
        forc.wind*log(zref / RSM.z0r) / log(parameter.windHeight / RSM.z0r) *
        log(zrUrb / UCM.z0u) / log(zref / UCM.z0u))
    dens = forc.pres / (1000 * 0.287042 * T_can * (1. + 1.607858 * UCM.canHum))

    # Friction velocity
    UCM.ustar = parameter.vk * windUrb / log((zrUrb - UCM.l_disp) / UCM.z0u)

    # Convective scaling velocity
    wstar = (parameter.g * max(UCM.sensHeat, 0.0) * zref / dens / Cp / T_can) ** (1 / 3.)
    UCM.ustarMod = max(UCM.ustar, wstar)  # Modified friction velocity
    UCM.uExch = parameter.exCoeff * UCM.ustarMod  # Exchange velocity

    # Canyon wind speed, Eq. 27 Chp. 3 Hanna and Britter, 2002
    # assuming CD = 1 and lambda_f = verToHor/4
    UCM.canWind = UCM.ustarMod * (UCM.verToHor / 8.) ** (-1 / 2.)

    # Canyon turbulent velocities
    UCM.turbU = 2.4 * UCM.ustarMod
    UCM.turbV = 1.9 * UCM.ustarMod
    UCM.turbW = 1.3 * UCM.ustarMod

    # Urban wind profile
    for iz in range(RSM.nzref):
        UCM.windProf.append(
            UCM.ustar / parameter.vk * log((RSM.z[iz] + UCM.bldHeight - UCM.l_disp) /
                                           UCM.z0u))

    return UCM, UBL, BEM
